#ifndef GENERATOR_SINGAUSS_HH
#define GENERATOR_SINGAUSS_HH

#include "DataSource.hh"

namespace generator {

  /**  SinGauss is an analytical function data source that generates a sinusoid
    *  of arbitrary phase multiplied by a gaussian envelope (i.e. a 
    *  sine-gaussian). The generated waveform is: 
    *  \begin{center}
    *  \[ x(t)  = A  \sin(\omega(t-t_0) + \Phi) e^{-(t-t_0)^2 / 2 \sigma^2} \]
    *
    *  $ \omega = 2 \pi F ; \;  \sigma = Q/\omega $ 
    *  \end{center}
    *
    *  The wave-form parameters that may be specified by the generation
    *  control function are:
    *
    *  \begin{tabular}{lcl}
    *  Parameter & Default & Meaning \\
    *  #A#     & - & Gaussian envelope amplitude at #t=to# \\
    *  #F#     & - & Sinusoid frequency (in Hz) \\
    *  #Q#     & - & Quality factor \\
    *  #Phi#   & 0 & Phase ($ \Phi $) at #t=to# (in degrees) \\
    *  #Width# & 6 & Half-width of the waveform to be generated (in $\sigma$
    *                units) \\
    *  \end{tabular} 
    *
    *  At the default width value of 6.0, the gaussian envelope has decreased 
    *  to $1.5 \times 10^{-8}$ of its central amplitude.
    *  @memo Sine-Gaussian function generator.
    *  @version 1.1; last modified April 21, 2004
    *  @author  John Zweizig
    */
  class SinGauss : public DataSource {
  public:
      /**  Default constructor
        */
      SinGauss(void);

      /**  Construct a SinGauss data source and initialize the waveform 
        *  parameters.
	*  @param Amp  the amplitude at #t=to#.
	*  @param Freq the sinusoid frequency (in Hz).
	*  @param Q    the quality factor.
	*  @param Phi  the sinusoid phase at #t=to# (in degrees).
	*  @param wid  the half width of the generated data in units of
	*              Q/(2*pi*F).
        */
      SinGauss(double Amp, double Freq, double Q, double Phi=0.0, 
	       double wid=6.0);

      /** Destroy a SinGauss data source.
        */
      ~SinGauss(void);

      /** Create an identical copy of a SinGauss object.
        */
      SinGauss* clone(void) const;

      /**  Return a constant string containing the data source type, "SinGauss"
        */
      const char* getDataType(void) const;

      /**  Return the duration of the data to be generated. For SinGauss this
        *  is evaluated on an event by event basis and is #2*Width*Q/F#.
        */
      Interval getDuration(void) const;

  private:
      /**  Calculate the event data series.
        */
      void getSeries(const Time& t0, Interval dT, int N, float* data);
  };

  //====================================  Inline methods.
  inline const char*
  SinGauss::getDataType(void) const {
      return "SinGauss";
  }

} // namespace generator

#endif  // GENERATOR_SINGAUSS_HH
