/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TRIG_SEG_ACCOUNTANT_HH
#define TRIG_SEG_ACCOUNTANT_HH

#include "Time.hh"
#include "Interval.hh"
#include <string>
#include <vector>

namespace trig {

    class Segment;
    class TrigClient;

    /**  The SegAccountant class performs segment accounting for a single
      *  segment generation process. Specifically, it:
      *  <ul>
      *    <li>Allows user to define multiple segments.</li>
      *    <li>Tracks segment status and whether it is defined.</li>
      *    <li>Floshes segment according to one of three different modes</li>
      *    <li> Allows user to define multiple segments.</li>>
      *  </ul>
      *
      *  Usage:
      *
      *  Initialization:
      *
      *  \verbatim
      *  SegAccountant sa(tc); // Construct
      *
      *  sa.addSegment("SgeName", 1, kAligned, 16.0);
      *  \endverbatim
      *
      *  Stride processing;
      *
      *  \verbatim
      *  sa.start_stride(ts.getStartTime());
      *  ...
      *  
      *  ...
      *  sa.update(ts.getEndTime());
      *  \endverbatim
      *
      *  \version 1.0; Last modified Jan 5, 2009
      *  \author  John G. zweizig
      */
    class SegAccountant {
    public:
	/** Enumerate the segment flush modes.
	 */
	enum flush_mode {
	    kNone,    // Allow segments to accumulate until the segment 
	              // changes state or until the accountant is deleted.
	    kMaxTime, // impose a maximum time period on all segments.
	    kAligned  // Flush at regular boundaries.
	};

	/**  Segment information structure.
	  */
	class seg_info {
	public:
	    /**  Construct a segment information entry for a named segment.
	      *  Online segments should have version 1.
	      *  @memo construct a segment information entry.
	      */
	    seg_info(const std::string& name, int version=1);

	    /**  Destroy a segment information entry.
	      *  Online segments should have version 1.
	      */
	    ~seg_info(void);

	    /**  Get the segment name.
	     */
	    const std::string& name(void) const;

	    /**  Get the last update time.
	     */
	    const Time& last(void) const;

	    /**  Set the segment flush mode and timing.
	     */
	    void set_mode(flush_mode m, Interval t=0);

	    /**  Set the segment state
	     */
	    void set_segment(const Time& start, const Time& end, bool onoff);

	    /**  Get the segment start time.
	     */
	    const Time& start(void) const;

	    /**  Get the segment name.
	     */
	    bool test(void) const;

	    /**  Get the segment name.
	     */
	    int version(void) const;

	private:
	    std::string _segname;
	    int         _version;
	    Time        _start;
	    Time        _last;
	    bool        _state;
	    flush_mode  _mode;
	    Interval    _delay;
	};

    public:
	/**  Construct an unattached segment accountant.
	  *  @memo Default constructor.
	  */
	SegAccountant(void);

	/**  Construct an unattached segment accountant.
	  *  @memo Default constructor.
	  */
	SegAccountant(TrigClient& tc);

	/**  Destroy an unattached segment accountant.
	  *  @memo Default constructor.
	  */
	~SegAccountant(void);

	/**  Add a segment to be accounted. The name is 
	 */
	void addSegment(const std::string& name, int version=1, 
			flush_mode mode=kNone, Interval dly=0.0);

	/**  Find the specified segment, return a reference.
	  */
	seg_info& find(const std::string& str);

	/**  Find the specified segment, return a reference.
	  */
	int locate(const std::string& str);

	/**  set up at start of analysis segment.
	  */
	void start_stride(const Time& ts);

	/**  Update to the specified time.
	  */
	void update(const Time& tup);

    private:
	typedef std::vector<seg_info>    seg_list;
	typedef seg_list::iterator       seg_iter;
	typedef seg_list::const_iterator const_seg_iter;

    private:
	seg_list    mList;
	TrigClient* mTC;
	Time        mSegStart;
    };

    //==================================  SegAccountant::seg_info inlines
    inline const std::string&
    SegAccountant::seg_info::name(void) const {
	return _segname;
    }

    inline const Time&
    SegAccountant::seg_info::last(void) const {
	return _last;
    }

    inline const Time& 
    SegAccountant::seg_info::start(void) const {
	return _start;
    }

    inline bool 
    SegAccountant::seg_info::test(void) const {
	return _state;
    }

    inline int 
    SegAccountant::seg_info::version(void) const {
	return _version;
    }
}

#endif  // !defined(TRIG_SEG_ACCOUNTANT_HH)
